--=====================================================================================
-- SKY BOX CHANGE SCRIPT
--
-- @Author:		akuenzi (akuenzi@gmail.com)
-- @Date:		2019-05-09
-- @Purpose:	To allow for different sky textures from time-to-time.
--=====================================================================================


SkyBoxChange = {}
SkyBoxChange.debug = false
SkyBoxChange.modDir = g_currentModDirectory;
SkyBoxChange.modsDir = g_modsDirectory;
SkyBoxChange.skiesDir = "FS17_SkySwap/SkyBoxes/"
SkyBoxChange.skyPackDir = "FS17_SkyPack/SkyBoxes/"

-- Skybox i3d files.  Mod map's skybox is added to table at first run.
SkyBoxChange.boxes =	{
					"SkyBoxes/GIANTS_FS09_SkyBox1.i3d",
					"SkyBoxes/GIANTS_FS09_SkyBox2.i3d",
					"SkyBoxes/GIANTS_FS11_SkyBox.i3d",
					"SkyBoxes/GIANTS_FS13_SkyBox.i3d",
					"SkyBoxes/GIANTS_FS13_SkyBox2.i3d",
					"SkyBoxes/GIANTS_FS15_SkyBox.i3d",
					"SkyBoxes/GIANTS_FS17_SkyBox1.i3d",
					"SkyBoxes/GIANTS_FS17_SkyBox2.i3d",
					"SkyBoxes/GIANTS_FS17_SkyBox3.i3d",
						};

-- Custom skybox textures.  Will be used to make random sky boxes.
SkyBoxChange.skies =	{
					
		day = 		{	-- Textures from modding sky packs.
						"Textures_Day/01_DAY_sky.dds",
						"Textures_Day/02_DAY_sky.dds",
						"Textures_Day/04_DAY_sky.dds",
						"Textures_Day/05_DAY_sky.dds",
						"Textures_Day/14_DAY_sky.dds",
						},
					
		night = 	{	-- Textures from modding sky packs.
						"Textures_Night/01_NIGHT_sky.dds",
						"Textures_Night/02_NIGHT_sky.dds",
						"Textures_Night/07_NIGHT_sky.dds",
						"Textures_Night/14_NIGHT_sky.dds",
						"Textures_Night/15_NIGHT_sky.dds",
						"Textures_Night/23_NIGHT_sky_LOF.dds",
						"Textures_Night/23_NIGHT_sky_TWC.dds",
						},
					
		overcast =	{	-- Textures from modding sky packs.
						"Textures_Overcast/01_OVERCAST_sky.dds",
						"Textures_Overcast/02_OVERCAST_sky.dds",
						"Textures_Overcast/05_OVERCAST_sky.dds",
						},
					
		sunrise =	{	-- Textures from modding sky packs.
						"Textures_Sunrise/01_SUNRISE_sky.dds",
						"Textures_Sunrise/02_SUNRISE_sky.dds",
						"Textures_Sunrise/05_SUNRISE_sky.dds",
						"Textures_Sunrise/07_SUNRISE_sky.dds",
						"Textures_Sunrise/14_SUNRISE_sky.dds",
						"Textures_Sunrise/15-2_SUNRISE_sky.dds",
						"Textures_Sunrise/18_SUNRISE_sky.dds",
						"Textures_Sunrise/20_SUNRISE_sky.dds",		
						},
					
		sunset =	{	-- Textures from modding sky packs.
						"Textures_Sunset/01_SUNSET_sky.dds",
						"Textures_Sunset/02_SUNSET_sky.dds",
						"Textures_Sunset/05_SUNSET_sky.dds",
						"Textures_Sunset/14_SUNSET_sky.dds",
						"Textures_Sunset/15-1_SUNSET_sky.dds",
						"Textures_Sunset/15-2_SUNSET_sky.dds",
						"Textures_Sunset/20_SUNSET_sky.dds",		
						},

						}

-- Define times of day.
SkyBoxChange.SUNRISE		= 1;
SkyBoxChange.DAY			= 2;
SkyBoxChange.SUNSET			= 3;
SkyBoxChange.OTHER			= 4;

-- Define maximum number of sky rotations in a day.
SkyBoxChange.ROTATIONS		= 10;

---------------------------------------------------------------------------------------------------------
-- STANDARD GAME FUNCTION OVERRIDES
---------------------------------------------------------------------------------------------------------

-- Save current skyBox being used to the career save game.
function SkyBoxChange:saveToXML()

	if self.isValid and self.xmlKey ~= nil then
		local key = self.xmlKey .. ".skyBox"
		removeXMLProperty(self.xmlFile, key);
		local mod = "SkyBoxChange";
		if _G[mod] ~= nil then
			_G[mod].saveBox(_G[mod], self.xmlFile, key)
		end;
	end;
	
end;
FSCareerMissionInfo.saveToXML = Utils.appendedFunction(FSCareerMissionInfo.saveToXML, SkyBoxChange.saveToXML);

function SkyBoxChange:saveBox(savegame, key)
	setXMLInt(savegame, key .. "#box", self.currentSky);
	setXMLString(savegame, key .. "#textures", self.textures);
end;

-- Load save game's skyBox that was in use during the last save.
function SkyBoxChange:loadFromXML()

	if g_currentMission == nil or not g_currentMission:getIsServer() then
		return;
	end;
	
	if g_currentMission.missionInfo.isValid then
		local xmlFile = g_currentMission.missionInfo.xmlFile;
		local key = g_currentMission.missionInfo.xmlKey .. ".skyBox";
		local mod = "SkyBoxChange";
		if _G[mod] ~= nil then
			_G[mod].load(_G[mod], xmlFile, key)	
		end;
	end;

end;
FSBaseMission.loadMapFinished = Utils.appendedFunction(FSBaseMission.loadMapFinished, SkyBoxChange.loadFromXML);

function SkyBoxChange:load(savegame, key)

	local currentSky = getXMLInt(savegame, key.."#box");
	self.textures = getXMLString(savegame, key.."#textures");
	if self.textures == nil then
		self.textures = "none"
	end;
	
	if self.startUp == nil then
		SkyBoxChange:startUpStuff();
	end;
	
	if currentSky ~= nil then
		self.currentSky = currentSky
	end;
	
	-- Need to check to see if the textures called for in the save game file still exist.
	if self.textures ~= "none" then
	
		-- Decompact saved data.
		local textures = SkyBoxChange.skies;
		local tex = Utils.splitString(" ", self.textures);
		local day = tonumber(tex[1]);
		local night = tonumber(tex[2]);
		local overcast = tonumber(tex[3]);
		local sunrise = tonumber(tex[4]);
		local sunset = tonumber(tex[5]);	
		
		-- Check to see if texture files exist.
		day = textures.day[day];
		night = textures.night[night];
		overcast = textures.overcast[overcast];
		sunrise = textures.sunrise[sunrise];
		sunset = textures.sunset[sunset];
		
		-- If any textures are not available, re-set to the map's original skybox.
		if day == nil or night == nil or overcast == nil or
			overcast == nil or sunrise == nil or sunset == nil then
			print("A texture file called for in the savegame file is missing.  Sky is reset to default.")
			self.currentSky = SkyBoxChange.NUM_SKIES - SkyBoxChange.NUM_TEXTURES
		end;
	end;
	
	if g_dedicatedServerInfo == nil then
		self:loadSky(self.currentSky);
	end;
end;

-- Add global stuff to facilitate multiplayer syncing.
function SkyBoxChange:loadStart()

	-- It appears readStream/writeStream doesn't work, and isn't called for a mod like this.
	-- We need to somehow get the game engine to call these functions...
	local mod = "SkyBoxChange";
	if _G[mod] ~= nil and _G[mod].preLoad ~= nil then
		_G[mod]:preLoad();
	end;

end;
Mission00.load = Utils.prependedFunction(Mission00.load, SkyBoxChange.loadStart)

---------------------------------------------------------------------------------------------------------
-- STANDARD MOD FUNCTIONS
---------------------------------------------------------------------------------------------------------

-- Stuff to do before mod loads up.  Mainly to tabulate texture and skybox info.
function SkyBoxChange:preLoad()
	-- Set up readStream/writeStream for multiplayer games.
	Server.sendObjects = Utils.prependedFunction(Server.sendObjects, skyJoinEvent.sendObjects);
	
	-- Update location of sky boxes in the event the zip file name changes.
	local file = SkyBoxChange.modDir .. "modDesc.xml"
	local name, _ = Utils.getModNameAndBaseDirectory(file)
	SkyBoxChange.skiesDir = tostring(name .. "/SkyBoxes/")
	
	if g_dedicatedServerInfo == nil then
		
		-- Copy generic skyBox to mods directory for later use.
		local modDir = SkyBoxChange.modDir
		local modsDir = SkyBoxChange.modsDir
		local boxFile = "OTHER_SkyBox01.i3d"
		local shapeFile = "OTHER_SkyBox01.i3d.shapes"
		
		-- copyFile(copyFile, destinationFile, overwrite boolean)
		copyFile(modDir .. "SkyBoxes/" .. boxFile, modsDir .. boxFile, true);
		copyFile(modDir .. "SkyBoxes/" .. shapeFile, modsDir .. shapeFile, true);
	end;
	
	-- Set up sky table stuff.
	-- Review pre-set skyboxes.
	local skyTable = SkyBoxChange.boxes;
	local boxes = {}
	for i, sky in ipairs(skyTable) do
		local fileCheck = Utils.getFilename(sky, SkyBoxChange.modDir)
		if not fileExists(fileCheck) then
			-- File does not exist, so don't include it in the sky box table.
			print("Sky box " .. sky .. " does not exist.")
		else
			-- File does exist.  Now update it for full file path.
			local j = #boxes + 1
			boxes[j] = fileCheck
		end;
	end;

	-- Replace existing table and count the number of skyboxes in it.
	SkyBoxChange.boxes = boxes
	SkyBoxChange.NUM_SKIES = table.getn(SkyBoxChange.boxes)
	
	-- Review random sky texture table.
	local skyTable = SkyBoxChange.skies
	for dayTime, textureType in pairs(skyTable) do
		local boxes = {}
		for j = 1, table.getn(textureType) do
			local sky = "SkyBoxes/" .. textureType[j]
			local fileCheck = Utils.getFilename(sky, SkyBoxChange.modDir)
			if not fileExists(fileCheck) then
				-- File does not exist, so remove it from the selection table.
				print("Sky swap texture file " .. textureType[j] .. " does not exist.")
			else
				-- File does exist.  Now update it for full file path.
				local name = string.lower(textureType[j]);
				if string.find(name, tostring(dayTime)) ~= nil then
					local i = #boxes + 1
					boxes[i] = fileCheck
				else
					print("Sky swap texture file " .. textureType[i] .. " needs to have the word '" .. tostring(dayTime) .. "' somewhere in its path/filename.")
				end;
			end;
		end;
		
		-- Replace existing table.
		SkyBoxChange.skies[dayTime] = boxes;
	end;

	-- Check to see if a sky pack is loaded to add textures from it.
	local skyPack = SkyBoxChange.skyPackDir .. "SkyPackInfo.xml";
	local filename = Utils.getFilename(skyPack, SkyBoxChange.modsDir);
	local packCount = 0
	if fileExists(filename) then
		local xmlFile = loadXMLFile("skyPackXML", filename);

		-- Grab additional textures and insert them in the sky table above.
		local i = 0;
		while true do
			local key = string.format("textures.File(%d)", i);
			if not hasXMLProperty(xmlFile, key) then
				break;
			end;
			local texName = getXMLString(xmlFile, key .. "#filename");
			local location = SkyBoxChange.skyPackDir .. texName
			local packName = Utils.getFilename(location, SkyBoxChange.modsDir);
			
			-- Check to see if the file exists.
			if not fileExists(packName) then
				print("Sky pack texture file " .. location .. " does not exist.")
			end;
			
			-- Insert texture into main table.
			local nameFile = string.lower(location);
			if nameFile ~= nil and fileExists(packName) then
				local badCount = 0
				if string.find(nameFile, "day") ~= nil then
					table.insert(SkyBoxChange.skies.day, packName);
				elseif string.find(nameFile, "night") ~= nil then
					table.insert(SkyBoxChange.skies.night, packName);
				elseif string.find(nameFile, "overcast") ~= nil then
					table.insert(SkyBoxChange.skies.overcast, packName);
				elseif string.find(nameFile, "sunrise") ~= nil then
					table.insert(SkyBoxChange.skies.sunrise, packName);
				elseif string.find(nameFile, "sunset") ~= nil then
					table.insert(SkyBoxChange.skies.sunset, packName);
				else
					-- If we end up here, the file exists, but it's not named properly.
					print("Sky pack texture file " .. location .. " needs to be properly named before it can be used in-game.  Its filename/path needs to have proper words included in it, such as 'day', 'night', 'overcast', 'sunrise', or 'sunset.'")
					badCount = -1
				end;
				packCount = packCount + badCount + 1;
			end;
			i = i + 1;
		end;

		delete(xmlFile);
	end;
	
	-- To determine maximum number of custom sky textures in the skies table.
	-- This dictates the number of 'other' skyboxes to use.
	local count = 0
	local check = false
	for _, sky in pairs(skyTable) do
		local amount = table.getn(sky);
		count = math.max(count, amount);
		if amount == 0 then
			check = true
		end;
	end;
	
	if not check then
		SkyBoxChange.NUM_TEXTURES = count;
	else
		SkyBoxChange.NUM_TEXTURES = 0;
		print("Generic sky box cannot be used until there is at least one each of day, night, overcast, sunrise, and sunset textures.")
	end;
	
end;

function SkyBoxChange:loadMap(name)
	self.firstRun = true;
	self.getSkyBoxData = SkyBoxChange.getSkyBoxData;
	self.timeOfDay = SkyBoxChange.timeOfDay;
	self.speed = SkyBoxChange.speed;
	self.rotateBox = SkyBoxChange.rotateBox;
	self.loadSky = SkyBoxChange.loadSky;
	self.loadMixedSky = SkyBoxChange.loadMixedSky;
	self.startUpStuff = SkyBoxChange.startUpStuff;

	self.swapTime = false														-- Boolean indicating when to swap out the skyBox.
	self.timer = 0																-- Timer for checking when to swap skyBox.
	self.skyRotateReset = false													-- Boolean indicating if the skyBox needs to be re-set or not.
	self.skySpeed = nil															-- Placeholder for speed of sky rotation.
	self.mapSize = 1															--[[Default map size.  Later it'll be multiplied by 1.42 to account for open space when a square shape (the sky box)
																					is rotated around another square shape (the map).]]
	self.textures = "none"														-- Placeholder for 'mixed' skybox textures.
	self.oldSky = nil;															-- Placeholder for old sky swapped out (requires fading).
	self.oldSkyTimer = nil;														-- Placeholder for old sky swapout timer.
	
	if self.startUp == nil then
		self:startUpStuff();
	end;

	--[[Add console command to activate/deactivate debug mode.
		It should be accessible only during single player mode.]]
	if not g_currentMission.missionDynamicInfo.isMultiplayer then
		addConsoleCommand("SkyBoxChangeDebugMode", "Sky box change mod debug mode.  Allows cycling through sky textures using input binding.", "consoleDebug", self);
	end;
	
	-- Add to environment listeners.
	g_currentMission.environment:addMinuteChangeListener(self)
	g_currentMission.environment:addDayChangeListener(self)

end;

function SkyBoxChange:deleteMap()
	--[[To remove the files put in the main mods directory to support this mod.
	Not able to be done since GIANTS deactivated the 'deleteFile()' command for mods.]]
	
	--[[local box = Utils.getFilename("OTHER_SkyBox01.i3d", SkyBoxChange.modsDir);
	local shape = Utils.getFilename("OTHER_SkyBox01.i3d.shapes", SkyBoxChange.modsDir);
	if fileExists(box) then
		deleteFile(box);
	end;
	if fileExists(shape) then
		deleteFile(shape);
	end;]]
	
	if not g_currentMission.missionDynamicInfo.isMultiplayer then
		removeConsoleCommand("SkyBoxChangeDebugMode");
	end;
	
end;

function SkyBoxChange:mouseEvent(posX, posY, modifier, isDown)
end;

function SkyBoxChange:keyEvent(unicode, sym, modifier, isDown)
end;

function SkyBoxChange:readStream(streamId, connection)
	self.currentSky = streamReadInt8(streamId);
	self.textures = streamReadString(streamId);
	self:loadSky(self.currentSky);
end;

function SkyBoxChange:writeStream(streamId, connection)
	streamWriteInt8(streamId, self.currentSky);
	streamWriteString(streamId, self.textures);
end;

function SkyBoxChange:update(dt)

	if self.firstRun then
		
		if g_dedicatedServerInfo == nil then
		
			-- Get skybox data at time of load, whether from savegame or initial load.
			self:getSkyBoxData();
			self.dayTime = self:timeOfDay()			-- Time of day (ie, SUNRISE, DAY, SUNSET, OR OTHER)
			
			-- Determine the map's size (for later stretching of the skyBox to accomodate map size).
			local size = g_currentMission.terrainSize
			if size >= 1024 * 2 * 4 then
				self.mapSize = 4 * 1.42 -- 16x map
			elseif size >= 1024 * 2 * 2 then
				self.mapSize = 2 * 1.42 -- 4x map
			end;
		
			-- Scale the opening sky.
			if g_currentMission:getIsClient() and self.mapSize ~= 1 then
				-- Scale the default-loaded sky (for 4x/16x maps) to ensure it's the proper size to rotate around the map.
				setScale(g_currentMission.environment.skyId, self.mapSize, self.mapSize, self.mapSize);
			end;
			
		end;
		
		self.firstRun = false
	end;

	-- Only do the following on client.
	if g_currentMission:getIsClient() and g_dedicatedServerInfo == nil then
		
		-- If we're in debug mode, then allow input binding to toggle the sky textures.
		if SkyBoxChange.debug then
			if InputBinding.hasEvent(InputBinding.Toggle) then
				self:rotateBox();
				if self.dayTime == SkyBoxChange.DAY then
					self.skySpeed = self:speed();
				end;
			end;
		end;
		
		-- Determine if it's time to swap out the skyBox or not.
		if self.swapTime then
			self:loadSky(self.currentSky)
			self.swapTime = false
		end;
		
		--Rotate sky when it's during the day time.
		if self.dayTime == SkyBoxChange.DAY then
			if self.skySpeed == nil then
				self.skySpeed = self:speed();
			end;
			local timeOfDay = g_currentMission.environment.dayTime								-- Time of current day in ms.
			local rotateProgress = (timeOfDay - self.dayTimeStartMS) / self.rotateTime			-- Percentage progress in rotation cycle.
			local rads = rotateProgress * self.skySpeed * (2 * math.pi)							-- Convert percentage of time period to radians.  2 * pi is a full circle.
			setRotation(g_currentMission.environment.skyId, 0, rads, 0);
			
			-- Set up need to re-set skyBox rotation once it's no longer day.
			if not self.skyRotateReset then
				self.skyRotateReset = true;
			end;
			
		else
			-- Reset skyBox position when it's NOT day time.
			if self.skyRotateReset then
				setRotation(g_currentMission.environment.skyId, 0, 0, 0);
				self.skyRotateReset = false;
				self.skySpeed = nil;
			end;
		end;
	end;
	
end;

function SkyBoxChange:draw()
end;

-- Additional 'stuff' to run upon startup
function SkyBoxChange:startUpStuff()

	-- Presumably this is called up on load if there's a save game, or on loadMap if a new game is started.

	--Save the map's existing sky to the table of skies.
	-- Locate the map's environment xml file where the sky i3d information is stored.	
	local xmlFilename = g_currentMission.environment.xmlFilename;
	local xmlFile = loadXMLFile("TempConfig", xmlFilename);
	local skyFile = Utils.getNoNil(getXMLString(xmlFile, "environment#filename"), "$data/sky/sky_day_night.i3d");
	skyFile = Utils.getFilename(skyFile, g_currentMission.baseDirectory)

	-- Insert the filename of the map's sky i3d file to this mod's table for later recall.
	table.insert(SkyBoxChange.boxes, skyFile);
	delete(xmlFile);
		
	--[[Add additional skies based on texture table count. These are the boxes selected 
		for randomized sky textures. This 'generic' box is placed in the main mods 
		directory at start up so it can be accessible while the game is running.  If it 
		was 'trapped' inside the zip file, then it couldn't be modified.]]
	local randomBox = "OTHER_SkyBox01.i3d";
	randomBox = Utils.getFilename(randomBox, SkyBoxChange.modsDir);
	if SkyBoxChange.NUM_TEXTURES ~= 0 then
		for i = 1, SkyBoxChange.NUM_TEXTURES do
			table.insert(SkyBoxChange.boxes, randomBox)
		end;
	end;
	
	-- Update number of entries in this mod's skies table.
	SkyBoxChange.NUM_SKIES = table.getn(SkyBoxChange.boxes)
		
	if self.currentSky == nil then
		-- This means we've just started a new savegame (or the sky box wasn't specified in the save game), and we don't know what sky we're using right now.
		self.currentSky = (SkyBoxChange.NUM_SKIES - SkyBoxChange.NUM_TEXTURES);
	end;

	-- Determine if the seasons mod is being used.
	if g_seasons ~= nil then
		g_seasons.daylight.dayChanged = Utils.appendedFunction(g_seasons.daylight.adaptTime, SkyBoxChange.adaptTime);
	end;
		
	-- Get skybox data at time of load, whether from savegame or initial load.
	if g_dedicatedServerInfo == nil then
		self:getSkyBoxData();
		self.dayTime = self:timeOfDay()			-- Time of day (ie, SUNRISE, DAY, SUNSET, OR OTHER)
		
		-- Set initial sky rotation at load time if it happens to be day.
		if self.dayTime == SkyBoxChange.DAY then
			if self.skySpeed == nil then
				self.skySpeed = self:speed();
			end;
			local timeOfDay = g_currentMission.environment.dayTime								-- Time of current day in ms.
			local rotateProgress = (timeOfDay - self.dayTimeStartMS) / self.rotateTime			-- Percentage progress in rotation cycle.
			local rads = rotateProgress * self.skySpeed * (2 * math.pi) 						-- Convert percentage of time period to radians.  2 * pi is a full circle.
			setRotation(g_currentMission.environment.skyId, 0, rads, 0);
		end;
	end;

	self.startUp = true
end;

-- Run this stuff every minute elapsed in-game.
function SkyBoxChange:minuteChanged()
	if g_dedicatedServerInfo == nil then
		self:timeOfDay();
	end;
end;

-- Run this stuff every time one day ends and transitions to the next.
function SkyBoxChange:dayChanged()
	if g_currentMission:getIsServer() and not SkyBoxChange.debug then
		self:rotateBox();
	end;
end;

-- To determine what 'period' of the day it is.  Mainly DAY or OTHER.
function SkyBoxChange:timeOfDay()

	local dayTime = g_currentMission.environment.dayTime			-- Time of current day in ms.
	local minute = dayTime / (1000 * 60)							-- Time of current day in minutes.
	
	if minute < self.dayTimeStart or minute > self.dayTimeEnd then
		self.dayTime = SkyBoxChange.OTHER;
		
	-- Check to see if we're in the day time hour (when skyBox can rotate).
	elseif minute >= self.dayTimeStart and minute <= self.dayTimeEnd then
		self.dayTime = SkyBoxChange.DAY;
	else
		self.dayTime = SkyBoxChange.OTHER;
	end;

end;

-- To get sky curve information from the sky box as to times when textures transition to/from DAY.
function SkyBoxChange:getSkyBoxData()
	
	local skyBox = g_currentMission.environment.skyCurve.keyframes
	local dayStart, dayEnd = 8*60, 17*60			-- Presumed day start/end times, to be modified by looking up actual data.
	local newDayTime1, newDayTime2
	
	for i = 1, table.getn(skyBox) do
	
		-- Find the time of day when it's on the 'day' skyBox setting (ie, only the day texture is showing).
		-- x = day time textures, y = sunset textures, z = night textures, w = sunrise textures
		if skyBox[i].x == 1 and skyBox[i].y == 0 and skyBox[i].z == 0 and skyBox[i].w == 0 then
			if newDayTime1 == nil then
				newDayTime1 = skyBox[i].time
			else
				newDayTime2 = skyBox[i].time
			end;
		end;
	end;
	
	-- Determine specific day start/end times.
	if newDayTime1 and newDayTime2 then
		dayStart = math.min(newDayTime1, newDayTime2);
		dayEnd = math.max(newDayTime1, newDayTime2);	
	end;

	-- Store these times for use later.
	self.dayTimeStart = dayStart				-- Time in minutes as to when sunrise hits and only day texture is showing in sky box (when sky can rotate).
	self.dayTimeEnd = dayEnd
	self.rotateTime = (self.dayTimeEnd - self.dayTimeStart) * 60 * 1000			-- Total time in ms during which sky may rotate.
	self.dayTimeStartMS = self.dayTimeStart * 60 * 1000							-- Time in ms at which sky may start to rotate.
	
	-- Debug info.
	if SkyBoxChange.debug then
		local hoursStart = math.floor(dayStart/60);
		local minutesStart = dayStart - (hoursStart * 60);
		local hoursEnd = math.floor(dayEnd/60);
		local minutesEnd = dayEnd - (hoursEnd * 60);
		print("Sky Box Times Updated:")
		print("Day time starts at " .. tostring(hoursStart) .. ":" .. tostring(minutesStart) .. " and ends at " .. tostring(hoursEnd) .. ":" .. tostring(minutesEnd));
		print("The period of time the sky will rotate is " .. tostring(self.dayTimeEnd-self.dayTimeStart) .. " minutes.");
	end;
	
end;

-- To determine number of random rotations of sky box each day.
function SkyBoxChange:speed()
	
	local filenameDay --,filenameOvercast
	if self.textures ~= "none" then
		local textures = SkyBoxChange.skies;
		local tex = Utils.splitString(" ", self.textures);
		local day = tonumber(tex[1]);
		--local overcast = tonumber(tex[3]);
		filenameDay = string.lower(textures.day[day]);
		--filenameOvercast = string.lower(textures.overcast[overcast]);
	end;

	-- Determine the number of times the sky will rotate during the day.	
	local speed = math.random(-SkyBoxChange.ROTATIONS, SkyBoxChange.ROTATIONS);
	if string.find(filenameDay, "norotate") ~= nil then --[[or string.find(filenameOvercast, "norotate") ~= nil]]
		speed = 0;
	end;
	
	-- Debug info.
	if SkyBoxChange.debug then
		print("Sky Box Daily Rotations Updated:")
		print("Number of Sky rotations each day -- " .. tostring(speed));
		if string.find(filenameDay, "norotate") ~= nil then
			print("Day sky texture " .. tostring(filenameDay) .. " is a non-rotating sky.");
		--elseif string.find(filenameOvercast, "norotate") ~= nil then
			--print("Overcast sky texture " .. tostring(filenameOvercast) .. " is a non-rotating sky.");
		end;
	end;
	
	return speed
	
end;

-- To set off process to grab a new sky box/textures.  This picks which box or which textures are involved.
function SkyBoxChange:rotateBox()
	-- Function only run on the server.
	
	-- Get new random sky.
	if SkyBoxChange.debug then
		-- To force use of custom skyBox when 'debug' is activated.
		self.currentSky = SkyBoxChange.NUM_SKIES
	else
		self.textures = "none";
		self.currentSky = math.random(1, SkyBoxChange.NUM_SKIES);
	end;
	
	-- Check to see if it's a 'swappable' box, and then pick new textures to show.
	if self.currentSky > (SkyBoxChange.NUM_SKIES - SkyBoxChange.NUM_TEXTURES) then
		
		-- Call table where texture information is stored.
		local textures = SkyBoxChange.skies;
		
		-- Pick random textures to use, and then call up related filenames.
		local day, night, overcast, sunrise, sunset
		if not SkyBoxChange.debug then
			-- What to do when we're not in debug mode.  Pick random textures.
			day = math.random(1, table.getn(textures.day));
			night = math.random(1, table.getn(textures.night));
			overcast = math.random(1, table.getn(textures.overcast));
			sunrise = math.random(1, table.getn(textures.sunrise));
			sunset = math.random(1, table.getn(textures.sunset));
		else
			-- What to do when we are in debug mode.  Pick textures sequentially.
			-- Below are the textures we're using currently plus 1 to get the next in sequence.
			local tex = Utils.splitString(" ", self.textures);
			day = tonumber(tex[1]) + 1;
			night = tonumber(tex[2]) + 1;
			overcast = tonumber(tex[3]) + 1;
			sunrise = tonumber(tex[4]) + 1;
			sunset = tonumber(tex[5]) + 1;
			
			-- Reset texture back to 1 if it's over the texture table total.
			if textures.day[day] == nil then day = 1 end;
			if textures.night[night] == nil then night = 1 end;
			if textures.overcast[overcast] == nil then overcast = 1 end;
			if textures.sunrise[sunrise] == nil then sunrise = 1 end;
			if textures.sunset[sunset] == nil then sunset = 1 end;
		end;
		
		-- Record numbers selected as a string so it can be passed through the event.
		self.textures = tostring(day) .. " " .. tostring(night) .. " " .. tostring(overcast) .. " " .. tostring(sunrise) .. " " .. tostring(sunset);
	
	end;
	
	self.swapTime = true
	
	-- Send information to the client.
	g_server:broadcastEvent(skyBoxEvent:new(self.currentSky, self.swapTime, self.textures), false);
	
end;

-- To load up a new sky box.
function SkyBoxChange:loadSky(skyId)
	
	if skyId ~= nil then
		-- Load up applicable skyBox file, if it exists.
		local skyTable = SkyBoxChange.boxes;
		
		if skyTable[skyId] ~= nil then
		
			-- Obtain file name of next sky box.
			local newSky = skyTable[skyId];
			if skyId > (SkyBoxChange.NUM_SKIES - SkyBoxChange.NUM_TEXTURES) and self.textures ~= "none" then
				-- One of the 'random' sky boxes has been selected, and we need to randomize the textures.
				newSky = self:loadMixedSky(skyId)
			end;
			
			if fileExists(newSky) then
				-- Load sky environment stored within this mod.
				newSky = loadI3DFile(newSky);
				local scale = self.mapSize;
				
				--[[Activate the new sky so that it shows up in the map.  This hopefully prevents the white 'blip' that
					can appear when the skyBox is just deleted cold.  The activation is temporary, as we've left the
					part showing to be a 'local' item.]]
				local child = getChildAt(newSky, 0);
					
				--[[Scale the size of the skyBox based on map size.  Skies on standard-sized maps are not scaled,
					as they are sufficiently large enough to rotate around already.]]
				if scale > 1 then
					setScale(child, scale, scale, scale);
				end;
					
				local x,y,z,w = g_currentMission.environment.skyCurve:get(g_currentMission.environment.dayTime / (1000*60))
				setShaderParameter(child, "partScale", x, y, z, w);
					
				-- Discard previous skyBox.
				delete(g_currentMission.environment.skyRootId);
				
				-- Establish the new skyBox, which is already showing.
				g_currentMission.environment.skyRootId = newSky;
				link(getRootNode(), g_currentMission.environment.skyRootId);
				g_currentMission.environment.skyId = getChildAt(g_currentMission.environment.skyRootId, 0);
				
				-- Scale the size of the skyBox based on map size.
				if scale > 1 then
					setScale(g_currentMission.environment.skyId, scale, scale, scale);
				end;			
				
			else
				print("SkyBoxChange mod is looking for " .. tostring(newSky) .. " but it does not exist.  Please check to see if it's included in the 'SkyBoxes' folder.");
			end;
		else
			if skyId ~= SkyBoxChange.NUM_SKIES then
				--[[Error message shouldn't show on multiplayer join, in which case the client-side map hasn't finished loading.  If the server map happens to be the last in the table,
					then it is the server map's sky, which hasn't yet been loaded by the client, but will be.  No error should show for this.]]
				print("SkyBoxChange mod is looking for sky # " .. tostring(skyId) .. ", but it does not exist.  Please check how many sky boxes are available in the 'SkyBoxChange.boxes' table.");
				
				-- Change current sky to last sky and re-run.
				self.currentSky = SkyBoxChange.NUM_SKIES
				self:loadSky(self.currentSky);
			end;
		end;
	end;
	
end;

-- When it's a generic sky box, this function is used to swap out textures for new one.
function SkyBoxChange:loadMixedSky(skyId)
	
	--Call tables where texture information is stored.
	local textures = SkyBoxChange.skies;
	
	-- Load texture indices previously determined.
	local tex = Utils.splitString(" ", self.textures);
	local day = tonumber(tex[1]);
	local night = tonumber(tex[2]);
	local overcast = tonumber(tex[3]);
	local sunrise = tonumber(tex[4]);
	local sunset = tonumber(tex[5]);

	-- Load texture file names based on the index.
	day = textures.day[day];
	night = textures.night[night];
	overcast = textures.overcast[overcast];
	sunrise = textures.sunrise[sunrise];
	sunset = textures.sunset[sunset];
		
	if SkyBoxChange.debug then
	--Debug Info (texture filename)
		print("CURRENT SKY")
		print("Day Texture = " .. tostring(day));
		print("Night Texture = " .. tostring(night));
		print("Overcast Texture = " .. tostring(overcast));
		print("Sunrise Texture = " .. tostring(sunrise));
		print("Sunset Texture = " .. tostring(sunset));
		print(" ");
	end;

	-- Load generic skybox i3d file as an xml file.	
	local name = SkyBoxChange.boxes[skyId];
	local xmlFile = loadXMLFile("skyNew", name);
	
	-- Swap out texture files in i3d file for new ones.
	local i = 0;
	while true do
		local key = string.format("i3D.Files.File(%d)", i);
		if not hasXMLProperty(xmlFile, key) then
			break;
		end;
		local nameFile = getXMLString(xmlFile, key .. "#filename");
		nameFile = string.lower(nameFile);
		if nameFile ~= nil then
			if string.find(nameFile, "day") ~= nil then
				setXMLString(xmlFile, key .. "#filename", day);
			elseif string.find(nameFile, "night") ~= nil then
				setXMLString(xmlFile, key .. "#filename", night);
			elseif string.find(nameFile, "overcast") ~= nil then
				setXMLString(xmlFile, key .. "#filename", overcast);
			elseif string.find(nameFile, "sunrise") ~= nil then
				setXMLString(xmlFile, key .. "#filename", sunrise);
			elseif string.find(nameFile, "sunset") ~= nil then
				setXMLString(xmlFile, key .. "#filename", sunset);
			elseif string.find(nameFile, "skyshader") ~= nil then
				-- We need to include this in case a user changes the name of the mod.
				setXMLString(xmlFile, key .. "#filename", SkyBoxChange.skiesDir .. "shaders/skyShader.xml");
			end;
		end;
		i = i+1;
	end;
	
	-- Save updated i3d file.
	saveXMLFile(xmlFile);
	delete(xmlFile);
	
	return name;
	
end;

-- If the seasons mod is present, this function forces update to sky box data.
function SkyBoxChange:adaptTime()
	-- After seasons changes have run, then update our sky box data.
	if g_dedicatedServerInfo == nil then
		SkyBoxChange:getSkyBoxData();
	end;
end;

-- Used for client syncing with server in multiplayer for updates to skyboxes or textures.
function SkyBoxChange:updateVariables(skyId, swapTime, textures)
	
	if not g_currentMission:getIsServer() then
		self.currentSky = skyId;
		self.swapTime = swapTime;
		self.textures = textures;
	end;

end;

-- To perform tasks required during debug mode when the console command is activated.
function SkyBoxChange:consoleDebug(toggle)
	
	if toggle == "true" then
		-- Turns on debug mode, which prints skydome texture file names to log.
		SkyBoxChange.debug = true
		
		print("SkyBoxChange mod is in debug mode.  Regular features will not work properly.")
		
		-- Switch to using only the custom sky dome.
		self.currentSky = SkyBoxChange.NUM_SKIES	-- Used to pick the custom skybox.
		self.swapTime = true						-- Toggle to ensure this sky is opened up.
				
		-- Start with opening textures in each category
		local day = 1
		local night = 1
		local overcast = 1
		local sunrise = 1
		local sunset = 1	
	
		-- Record numbers selected as a string for sky loadup.
		self.textures = tostring(day) .. " " .. tostring(night) .. " " .. tostring(overcast) .. " " .. tostring(sunrise) .. " " .. tostring(sunset);		
		
		-- Update speed with a new sky.
		if self.dayTime == SkyBoxChange.DAY then
			self.skySpeed = self:speed();
		end;
		
		-- Indicate if seasons mod is present.
		if g_seasons ~= nil then
			print("Season's mod detected.  Additional code appended to adapt sky box changes to seasonal changes.")
		end;
		
		-- Indicate the size of the map according to this mod.
		local x = 1								-- Default is standard-sized map.
		if self.mapSize >= 4 * 1.42 then 		-- 16x map
			x = 16 
		elseif self.mapSize >= 2 * 1.42 then	-- 4x map
			x = 4
		end;
		print("General sky box change debug information:")
		print("The map size appears to be " .. tostring(x) .. "x.")
		print("The skybox i3d file x,y,z settings have been scaled by " .. tostring(self.mapSize) .. ".");
		
		-- Indicate the number of textures and sky boxes being used.
		local boxes = SkyBoxChange.boxes
		local unique = 0									-- This counts up the number of skyboxes being used.
		for i = 1, table.getn(boxes) do
			if boxes[i] ~= boxes[i-1] then
				unique = unique + 1
			end;
		end;
		print("The skybox change mod will choose between " .. tostring(SkyBoxChange.NUM_SKIES) .. " sky boxes each day.")
		print("These consist of one original sky box (the one that came with the map), " .. tostring(unique-2) .. " additional unique sky boxes, and one generic sky box for the remaining " .. tostring(SkyBoxChange.NUM_SKIES - unique + 1) .. " slots.")
		
		print("When selected, the generic sky box will choose/swap between:")
		local boxes = SkyBoxChange.skies
		for textureName, textureType in pairs(boxes) do
			local textureCount = table.getn(textureType)	-- This counts up the number of sky textures being used.
			print("     " .. tostring(textureCount) .. " different " .. tostring(textureName) .. " textures.")
		end;

	elseif toggle == "skies" then
		--Prints entire set of sky box and sky texture filenames to log.
		local boxes = SkyBoxChange.boxes
		for i = 1, table.getn(boxes) do
			print("[" .. tostring(i) .. "] - " .. tostring(boxes[i]))
		end;
		
		boxes = SkyBoxChange.skies
		for _, textureType in pairs(boxes) do
			for i = 1, table.getn(textureType) do
				print("[" .. tostring(i) .. "] - " .. tostring(textureType[i]))
			end;
		end;
	elseif tonumber(toggle) <= (SkyBoxChange.NUM_SKIES - SkyBoxChange.NUM_TEXTURES) then
		-- Load custom skyBox by #.
		toggle = tonumber(toggle)
		self.currentSky = toggle;			-- Toggle to open noted sky box.
		self.swapTime = true;				-- Toggle to ensure this sky is opened up.
		print("CURRENT SKY")
		print("Sky Box # = " .. tostring(toggle));
		print("Sky Box Filename = " .. tostring(SkyBoxChange.boxes[toggle]));
	else
		-- Anything else turns the debug mode off.
		SkyBoxChange.debug = false
	end;

end;

addModEventListener(SkyBoxChange);


---------------------------------------------------------------------------------------------------------
--  SET SKY EVENT FOR MULTIPLAYER	(ie, WHEN SKY CHANGES ON THE SERVER, CHANGE IT ON THE CLIENT)
---------------------------------------------------------------------------------------------------------
skyBoxEvent = {};
skyBoxEvent_mt = Class(skyBoxEvent, Event);

InitEventClass(skyBoxEvent, "skyBoxEvent");

function skyBoxEvent:emptyNew()
	local self = Event:new(skyBoxEvent_mt);
	self.className = "skyBoxEvent";
	return self;
end;

function skyBoxEvent:new(skyId, swapTime, textures)
	local self = skyBoxEvent:emptyNew()
	self.currentSky = skyId;
	self.swapTime = swapTime;
	self.textures = textures;
	return self;
end;

function skyBoxEvent:readStream(streamId, connection)
	self.currentSky = streamReadInt8(streamId);
	self.swapTime = streamReadBool(streamId);
	self.textures = streamReadString(streamId);
	SkyBoxChange:updateVariables(self.currentSky, self.swapTime, self.textures);
end;

function skyBoxEvent:writeStream(streamId, connection)
	streamWriteInt8(streamId, self.currentSky);
	streamWriteBool(streamId, self.swapTime);
	streamWriteString(streamId, self.textures);
end;

function skyBoxEvent:run(connection)
	--print("Nothing to run here as the server isn't receiving information from clients.");
end;


---------------------------------------------------------------------------------------------------------
--  MULTIPLAYER JOIN EVENT (ie, UPON JOIN, SET UP INITIAL MP SYNCING)
---------------------------------------------------------------------------------------------------------
skyJoinEvent = {}
skyJoinEvent_mt = Class(skyJoinEvent, Event)

InitEventClass(skyJoinEvent, "skyJoinEvent")

function skyJoinEvent:emptyNew()
    local self = Event:new(skyJoinEvent_mt);
    self.className = "skyJoinEvent";
    return self
end;

function skyJoinEvent:new()
    local self = skyJoinEvent:emptyNew();
    return self;
end;

function skyJoinEvent:writeStream(streamId, connection)
    if not connection:getIsServer() then
		local mod = "SkyBoxChange";
		if _G[mod].writeStream ~= nil then
			_G[mod]:writeStream(streamId, connection)
		end;
    end;
end;

function skyJoinEvent:readStream(streamId, connection)
    if connection:getIsServer() then
		local mod = "SkyBoxChange";
		if _G[mod].readStream ~= nil then
			_G[mod]:readStream(streamId, connection)
		end;
    end;
end;

function skyJoinEvent:sendObjects(connection, x, y, z, viewDistanceCoeff)
    connection:sendEvent(skyJoinEvent:new())
end;